
# シンプルなCNN [normal_cnn]

import sys
import numpy as np

# 標準的な畳み込み層
def normal_conv_layer(x, n_filter, kernel_size):
	from keras import layers
	x = layers.Conv2D(n_filter, kernel_size, padding = "same")(x)
	x = layers.BatchNormalization()(x)
	x = layers.Activation("relu")(x)
	return x

# skip-connection有りの畳み込み層
def residual_layer(x, n_filter, kernel_size):
	from keras import layers
	prev_x = x # skip_connection用に退避
	x = layers.Conv2D(n_filter, kernel_size, padding = "same")(x)
	x = layers.BatchNormalization()(x)
	x = layers.Activation("relu")(x)
	x = layers.Conv2D(n_filter, kernel_size, padding = "same")(x)
	x = layers.BatchNormalization()(x)
	x = layers.add([x, prev_x])	# 合流部
	x = layers.Activation("relu")(x)
	return x

# 全結合層
def dense_layer(x, unit_n, activation = None):
	from keras import layers
	x = layers.Dense(unit_n, activation = activation)(x)
	return x

# シンプルなCNN [normal_cnn]
class Normal_CNN():
	# 初期化処理
	def __init__(self, n_out, nn_param, loss = "mse", optimizer = "sgd"):
		self.keras_model = None
		self.n_out = n_out
		self.nn_param = nn_param
		self.loss = loss
		self.optimizer = optimizer
	# 学習
	def fit(self, input_ls, output_ls, epochs = 10):
		if self.keras_model is None:
			image_shape = input_ls[0].shape
			# keras-cnnの初期化
			ret_nn = self.instantiate_nn(input_shape = image_shape)
			self.keras_model = ret_nn
		self.keras_model.fit(input_ls, output_ls, epochs = epochs)
	# サンプル中の1つのバッチで勾配を更新
	def train_on_batch(self, input_ls, output_ls):
		if self.keras_model is None:
			image_shape = input_ls[0].shape
			# keras-cnnの初期化
			ret_nn = self.instantiate_nn(input_shape = image_shape)
			self.keras_model = ret_nn
		loss_value = self.keras_model.train_on_batch(input_ls, output_ls)
		return loss_value
	# 予測
	def predict(self, input_ls):
		pred_ls = self.keras_model.predict(input_ls)
		return pred_ls
	# keras-cnnの初期化
	def instantiate_nn(self, input_shape):
		from keras import layers
		# 入力層の定義
		inputs = layers.Input(shape = input_shape)
		# 畳み込み層の定義
		feature = inputs
		for one_info in self.nn_param["conv_layer_info"]:
			n_filter = one_info["n_filter"]
			kernel_size = one_info["kernel_size"]
			feature = normal_conv_layer(feature, n_filter, kernel_size)	# skip-connection有りの畳み込み層
		# Flatten
		feature = layers.Flatten()(feature)
		# 全結合層、出力層の定義
		for one_info in self.nn_param["dense_layer_info"]:
			unit_n = one_info["unit_n"]
			activation = one_info["activation"]
			feature = dense_layer(feature, unit_n = unit_n, activation = activation)	# 全結合層
		output_activation = self.nn_param["output_activation"]
		outputs = dense_layer(feature, unit_n = self.n_out, activation = output_activation)	# 出力層
		# 層をまとめる
		from keras.models import Model
		model = Model(inputs = inputs, outputs = outputs)
		model.compile(loss = self.loss, optimizer = self.optimizer)
		return model




import sys
from sout import sout
# 共通部品 [utils.py]
from .parts.utils import TileArt
# デバッグ用のプレーヤー [debug_players.py]
from .parts.debug_players import RandomPlayer, HumanPlayer
# Deep Q Network AI [DQN_Agent]
from .parts.DQN_Agent import DQN_Agent
# Catcherゲーム [catcher_game]
from .parts import catcher_game
# # 紅白ゲーム [rw_game]
# from .parts import rw_game
# デモ [demo.py]
from .parts.demo import demo
# エピソード関係処理 [episode.py]
from .parts.episode import do_episode, do_episodes
# 棋譜の画像化 (デフォルトは入力棋譜と同じディレクトリに置かれる)
from .parts.history import history2img




# Deep Q Network AI for testing [test_agent.py]

import sys
# DQNの行動決定 (eps greedy戦略) [policy.py]
from .policy import eps_greedy_policy

# Deep Q Network AI for testing [test_agent.py]
class DQN_Test_Agent:
	# 初期化処理
	def __init__(self, ai_obs, qnet, action_ls, eps):
		# 各種パラメータ
		self.ai_obs, self.action_ls, self.eps = ai_obs, action_ls, eps
		# この時点のqnetをコピー
		self.fixed_qnet = qnet.clone_model(	# 複製体を作る (返り値はNormal_CNN型) [normal_cnn]
			untrained_check = False)
	# 行動決定
	def think(self, state, reward):
		obs_img = self.ai_obs(state)
		selected_a = eps_greedy_policy(obs_img, self.fixed_qnet, self.action_ls, self.eps)	# DQNの行動決定 (eps greedy戦略) [policy.py]
		if state["finished"] is True: selected_a = "finish_action"
		return selected_a	# 選択したactionを返す




# qnetの学習 [train_qnet.py]

import sys
import random
import numpy as np

# 攪拌 + リサンプリングするためのインデックスのリストを生成
def gen_random_idx_ls(n, ratio):
	idx_ls = random.sample(list(range(n)), k = int(n * ratio))
	return idx_ls

# playlogから教師データの生成 [train_qnet.py]
def gen_train_data(playlog, tnet, action_ls, dqn_param):
	# 全レコードを取得 [playlog.py]
	playlog_data = playlog.get_all_data()
	# prev_obsが存在しないデータを除く
	filtered_playlog_ls = [e for e in playlog_data if e.get("prev_obs", None) is not None]
	# 学習データ(入力側)の生成
	input_ls = np.array([one_log["prev_obs"] for one_log in filtered_playlog_ls])
	# 学習データ(出力側)の生成
	esitimated_ls = tnet.predict(input_ls)	# 学習の安定のため、重みを固定したNNによってQ値を計算
	output_ls = np.array(esitimated_ls)
	action_idx_dict = {a:i for i, a in enumerate(action_ls)} # アクション名から通し番号を取り出す辞書
	for log_idx, one_log in enumerate(filtered_playlog_ls):
		# エピソード終了の1つ前のstateのみ遷移先基準のQ値が計算できないため、条件分岐
		if one_log["now_finished"] is True:
			gain = one_log["now_reward"]
		else:
			if log_idx + 1 >= len(esitimated_ls): continue	# 次stepの情報がないやつは学習に加えない
			next_s_max_q = max(esitimated_ls[log_idx + 1])
			gain = one_log["now_reward"] + dqn_param["gamma"] * next_s_max_q
		# 実際に取った行動のQ値のみ更新
		action_idx = action_idx_dict[one_log["prev_action"]]	# アクション名から通し番号を取り出す辞書
		output_ls[log_idx][action_idx] = gain
	# 攪拌 + リサンプリングして学習
	idx_ls = gen_random_idx_ls(n = len(input_ls), ratio = dqn_param["train_data_ratio"])	# 攪拌 + リサンプリングするためのインデックスのリストを生成
	train_x, train_y = input_ls[idx_ls], output_ls[idx_ls]
	return train_x, train_y

# qnetの学習 [train_qnet.py]
def train_qnet(qnet, tnet, playlog, action_ls, dqn_param):
	# playlogから教師データの生成 [train_qnet.py]
	train_x, train_y = gen_train_data(playlog, tnet, action_ls, dqn_param)
	# 学習データが0件のときに学習をスキップする
	if len(train_x) == 0:
		print("[DQN_Agent warning] gen_train_data()の結果生成された学習でーたが0件のため、学習をスキップします。")
		return None
	# サンプル中の1つのバッチで勾配を更新 [normal_cnn]
	loss = qnet.train_on_batch(train_x, train_y)
	# print("loss: %.4f"%loss)




import sys
import copy

###### resoutに移行したため廃止 (画面のない環境に対応するため)
# # 画像の表示
# def show_img(img, title_msg = "show_img", waitKey = True, ratio = 1.0, interpolation = "INTER_CUBIC"):
# 	import cv2
# 	interpolation_dic = {"INTER_CUBIC": cv2.INTER_CUBIC, "INTER_NEAREST": cv2.INTER_NEAREST}
# 	if interpolation not in interpolation_dic: raise Exception("[ezRL error] 未対応のinterpolation値が指定されました")
# 	resized_img = cv2.resize(img, None, fx = ratio, fy = ratio, interpolation = interpolation_dic[interpolation])
# 	cv2.imshow(title_msg, resized_img)
# 	# キー入力待機
# 	if waitKey is not False:
# 		if waitKey is True: waitKey = 0
# 		if type(waitKey) != type(0): raise Exception("[ezRL error] show_img()のwaitKey引数は整数または真理値である必要があります。")
# 		cv2.waitKey(waitKey)

# 画像を透過して合成
def transparent_mix(base_img, mix_image, transparent_color):
	import numpy as np
	# 真っ黒の画像を作成
	gen_black = lambda: np.zeros(base_img.shape, dtype = np.uint8)
	# 画像のコピー
	cp = copy.deepcopy
	# 透過マスクの作成
	tp = np.array(transparent_color, dtype = np.uint8)[None, None, :]
	raw_mask = gen_black()
	raw_mask[mix_image == tp] = 1
	sum_mask = np.sum(raw_mask, axis = 2)
	mask_2d = (sum_mask == 3)
	mask = (np.tile(mask_2d[:,:,None], (1,1,3)) == False)
	# 合成
	img_0, img_1 = cp(base_img), cp(mix_image)
	img_0[mask] = 0
	img_1[mask == False] = 0
	mixed_img = img_0 + img_1
	return mixed_img

# タイルを並べた画像の作成 [ezRL]
class TileArt:
	# 初期化処理
	def __init__(self,
		canvas_size,	# キャンバスの大きさ
		tile_file_dic,	# タイル名と画像ファイルパスの対応辞書
		transparent_color = (255, 0, 255),	# 透過色の指定 (b,g,r)
		window_title = "TileArt"
	):
		import cv2
		# 画像の読み込み
		self.tile_img_dic = {
			tile_name: cv2.imread(tile_file_dic[tile_name])	# 画像読み込み
			for tile_name in tile_file_dic
		}
		# 様々なパラメータを保持
		self.canvas_size = canvas_size	# キャンバスの大きさ
		self.transparent_color = transparent_color	# 透過色の指定 (b,g,r)
		self.window_title = window_title
	# タイルを並べた画像の作成
	def gen_img(self, tile_ls, show_flag = False, ratio = 1.0, interpolation = None):
		import cv2
		import numpy as np
		# 補完のデフォルト値
		if interpolation is None: interpolation = cv2.INTER_NEAREST
		# キャンバスの作成
		x_canvas_size, y_canvas_size = self.canvas_size
		img = np.zeros((y_canvas_size, x_canvas_size, 3), dtype = np.uint8)
		# layer順にソート (安定ソート)
		sorted_tile_ls = sorted(tile_ls, key = lambda e: e["layer"])
		# タイルを並べる
		for tile in sorted_tile_ls:
			px, py = tile["pos"]
			x_size, y_size = tile["size"]
			texture = tile["texture"]
			raw_tile_img = self.tile_img_dic[texture]
			# 画像を伸縮する
			tile_img = cv2.resize(raw_tile_img, (x_size, y_size), interpolation = interpolation)
			# 画像を透過して合成
			base_img = img[py:py+y_size,px:px+x_size,:]
			mixed_img = transparent_mix(base_img, tile_img, self.transparent_color)	# 画像を透過して合成
			img[py:py+y_size,px:px+x_size,:] = mixed_img
		# 画像の表示 (show_flagがTrueの時)
		if show_flag is True: show_img(img, ratio)	# 画像の表示
		# 出来上がった画像を返す
		return img




# Deep Q Network AI [DQN_Agent]
# 【動作確認・使用例】

import sys
import resout as rout
import matplotlib.pyplot as plt
from relpath import add_import_path, rel2abs
add_import_path("../")
# Catcherゲーム [catcher_game]
from catcher_game import game_step, all_actions, human_obs, ai_obs, gen_init_state
# Deep Q Network AI [DQN_Agent]
from DQN_Agent import DQN_Agent

# 保存パスの設定 [resout]
rout.set_save_dir("./output_img/")

train_ai = DQN_Agent(action_ls = all_actions({}), ai_obs = ai_obs) # Deep Q Network AI
episode_n = 700	# エピソード数
total_reward_ls = []
for episode_idx in range(episode_n):
	# state, actionの初期化
	action, state = "initial_action", gen_init_state(game_params = {})
	reward_ls = []
	# ゲーム進行
	while state["finished"] is False:
		state, reward = game_step(state, action)
		action = train_ai.think(state, reward) # 行動決定
		reward_ls.append(reward)
	total_reward_ls.append(sum(reward_ls))
	print("Episode #%d, Reward: %.1f"%(episode_idx, total_reward_ls[-1]))
# 獲得報酬の推移を表示
plt.plot(total_reward_ls)
plt.savefig(rout.gen_save_path(".png"))	# 保存ファイル名の生成(自動で連番になる) [resout]
# テスト用プレーヤーを生成 [DQN_Agent]
test_ai = train_ai.gen_test()
# state, actionの初期化
action, state = "initial_action", gen_init_state(game_params = {})
# テストプレイ
while state["finished"] is False:
	img = human_obs(state)	# 人間用の観測関数
	rout.save_img(img, ratio = 2)	# 画像の保存 [resout]
	state, reward = game_step(state, action)
	action = test_ai.think(state, reward)




# Deep Q Network AI [DQN_Agent]

import sys
# Triggerオブジェクト [trigger.py]
from .parts.trigger import Trigger
# eps値のスケジュール管理 [utils.py]
from .parts.utils import EpsAdmin
# プレイ記録 [playlog.py]
from .parts.playlog import DQNPlaylog
# 未学習時における予測関数の生成
from .parts.utils import gen_rand_pred_func
# qnetの双子 (qnetと重みを時々同期する) [target_twin.py]
from .parts.target_twin import TargetTwin
# DQNの行動決定 (eps greedy戦略) [policy.py]
from .parts.policy import eps_greedy_policy
# qnetの学習 [train_qnet.py]
from .parts.train_qnet import train_qnet
# Deep Q Network AI for testing [test_agent.py]
from .parts.test_agent import DQN_Test_Agent
from relpath import add_import_path
add_import_path("../")
# シンプルなCNN [normal_cnn]
from normal_cnn import Normal_CNN

# qnetのネットワーク構成 (default構成)
default_nn_param = {
	"conv_layer_info": [
		{"n_filter": 8, "kernel_size": (3, 3)},
		{"n_filter": 16, "kernel_size": (3, 3)},
		# {"n_filter": 32, "kernel_size": (5, 5)},
	],
	"dense_layer_info": [
		{"unit_n": 512, "activation": "tanh"},
		{"unit_n": 128, "activation": "tanh"},
		{"unit_n": 32, "activation": "tanh"},
		{"unit_n": 8, "activation": "tanh"},
	],
	"output_activation": None
}

# Deep Q Network AI
class DQN_Agent:
	# 初期化処理
	def __init__(self,
		action_ls, ai_obs, gamma = 0.9, buffer_size = 5000,
		train_data_ratio = 0.1, # Experience Replay時にlogから取り出すデータの割合
		learning_trigger = ["step_n", 100],	# qnet学習のタイミング (["finished_n", 3], ["step_n", 300], 関数指定など)
		target_update_trigger = ["step_n", 300],	# target_net更新のタイミング (["finished_n", 3], ["step_n", 300], 関数指定など)
		eps_schedule = [0.5, 0.0, "finished_n", 500],	# epsの変化スケジュール ([start, end, n_type, n]または関数指定)
		nn_param = default_nn_param
	):
		self.action_ls = action_ls
		self.ai_obs = ai_obs	# ai_obs_imgを生成する関数
		self.dqn_param = {"gamma": gamma, "train_data_ratio": train_data_ratio}	# dqnの学習等設定パラメータ
		self.learning_trigger = Trigger(learning_trigger)	# Triggerオブジェクトの初期化 (指定方法: ["finished", 3]または関数指定) [trigger.py]
		self.target_update_trigger = Trigger(target_update_trigger)	# Triggerオブジェクトの初期化 (指定方法: ["finished", 3]または関数指定) [trigger.py]
		self.eps_admin = EpsAdmin(eps_schedule)	# eps値のスケジュール管理 [utils.py]
		self.playlog = DQNPlaylog(buffer_size)	# プレイ記録 [playlog.py]
		rand_pred_func = gen_rand_pred_func(len(self.action_ls))	# 未学習時における予測関数の生成 (numpy配列の型で生成)
		self.qnet = Normal_CNN(len(self.action_ls), nn_param, loss = "mse", optimizer = "RMSprop", untrained_pred_func = rand_pred_func)	# シンプルなCNN [normal_cnn]
		self.tnet = TargetTwin(self.qnet)	# qnetの双子 (qnetと重みを時々同期する) [target_twin.py]
		self.step_cnt, self.finished_cnt = 0, 0	# アクションのステップ数・finishedを通過した回数 (多くの場合エピソードの回数に対応)
	# 行動決定
	def think(self, state, reward):
		obs_img = self.ai_obs(state)
		selected_a = eps_greedy_policy(obs_img, self.qnet, self.action_ls, self.eps_admin.get_eps())	# DQNの行動決定 (eps greedy戦略) [policy.py]
		if state["finished"] is True: selected_a = "finish_action"
		self.playlog.append(selected_a, state, obs_img, reward)	# playlog追記 [playlog.py]
		self.__train_and_update(state["finished"])	# 学習や状態更新 (finished_cntの更新など)
		return selected_a	# 選択したactionを返す
	# 学習や状態更新 (finished_cntの更新など)
	def __train_and_update(self, finished_flag):
		# カウンターのインクリメント
		self.step_cnt += 1
		if finished_flag is True: self.finished_cnt += 1
		# モデルの学習
		if self.learning_trigger.judge(self.step_cnt, self.finished_cnt):	# triggerタイミングの判断 [trigger.py]
			train_qnet(self.qnet, self.tnet, self.playlog, self.action_ls, self.dqn_param)	# qnetの学習 [train_qnet.py]
		# tnet(target)の更新
		if self.target_update_trigger.judge(self.step_cnt, self.finished_cnt):	# triggerタイミングの判断 [trigger.py]
			self.tnet.update_by_twin_weight()	# weightを更新 (双子の姉のweightに基づく) [target_twin.py]
		# EpsAdminの更新
		self.eps_admin.update(self.step_cnt, self.finished_cnt)	# eps値算出根拠の更新
	# テスト用プレーヤーを生成 [DQN_Agent]
	def gen_test(self):
		# Deep Q Network AI for testing [test_agent.py]
		test_ai = DQN_Test_Agent(self.ai_obs, self.qnet, self.action_ls, eps = 0.0)
		return test_ai




# 共通処理群 [normal_cnn]

import sys

# デフォルトの「未学習時における予測関数」 (エラーで落ちる挙動)
def default_untrained_pred_func(test_x):
	raise Exception("[normal_cnn error] 未学習の状態でprediction()関数が呼ばれました (未学習時の挙動は、NormalCNNクラス初期化時の引数untrained_pred_funcにより指定することができます。)")




# エピソード関係処理 [episode.py]

import sys
import copy
import json
import resout as rout
import matplotlib.pyplot as plt
# 棋譜をjson形式で保存 [history.py]
from .history import save_history_json

# エピソードを実行 [episode.py]
def do_episode(game, ai, game_params, save_history = False, episode_label = None):
	# defaultのepisode_label
	if episode_label is None: episode_label = "no_name_episode"
	# state, actionの初期化
	action, state = "initial_action", game.gen_init_state(game_params)
	# 棋譜追記
	history_ls = []
	def add_history(state, action, reward):
		if save_history is True: state = copy.deepcopy(state)
		history_ls.append({"state": state, "action": action, "reward": reward})
	# ゲーム進行
	add_history(state, action, reward = None)	# 棋譜追記
	while state["finished"] is False:
		state, reward = game.game_step(state, action)
		action = ai.think(state, reward) # 行動決定
		add_history(state, action, reward)	# 棋譜追記
	# 棋譜の保存
	if save_history is True:
		save_history_json(history_ls, rout.gen_save_path(".json", label = episode_label))	# 棋譜をjson形式で保存 [history.py]
	return history_ls

# 棋譜から報酬合計値を算出
def reward_sum(history_ls):
	# reward一覧
	reward_ls = [step["reward"] for step in history_ls]
	# Noneを除く
	non_none_r_ls = [r for r in reward_ls if r is not None]
	# 合計を返す
	return sum(non_none_r_ls)

# episode結果・進捗の表示
def show_episode_result(episode_idx, episode_label, reward):
	if episode_label is None:
		episode_label_str = ""
	else:
		episode_label_str = " (%s)"%episode_label
	print("Episode #%d%s, Reward: %.1f"%(episode_idx, episode_label_str, reward))

# 複数エピソード実行 [episode.py]
def do_episodes(game, train_ai, game_params_ls, episode_label_ls = None, callback = None, save_history = False, save_reward_ls = False):
	if episode_label_ls is None: episode_label_ls = [None for _ in game_params_ls]
	total_reward_ls = []
	for episode_idx, game_params in enumerate(game_params_ls):
		history_ls = do_episode(game, train_ai, game_params, save_history, episode_label = episode_label_ls[episode_idx])	# エピソードを実行
		total_reward_ls.append(reward_sum(history_ls))	# 棋譜から報酬合計値を算出
		show_episode_result(episode_idx, episode_label_ls[episode_idx], total_reward_ls[-1])	# episode結果・進捗の表示
		# コールバック関数の実行
		if callback is not None: callback(train_ai, episode_idx)
	# 獲得報酬の推移を表示
	if save_reward_ls is True:
		plt.plot(total_reward_ls)
		plt.savefig(rout.gen_save_path(".png"))	# 保存ファイル名の生成(自動で連番になる) [resout]




# 棋譜関係処理 [history.py]

import os
import sys
import json
import resout as rout

# 棋譜をjson形式で保存 [history.py]
def save_history_json(history_ls, filename):
	# stepごとに改行区切りしたjsonの作成
	step_ls = [
		"  " + json.dumps(step, ensure_ascii = False)
		for step in history_ls]
	comma_ls = ",\n".join(step_ls)
	json_str = "[\n%s\n]"%comma_ls
	# 保存ファイル名の生成(自動で連番になる) [resout]
	with open(filename, "w", encoding = "utf-8") as f:
		f.write(json_str)

# 保存先フォルダ作成
def make_save_dir(history_filename):
	ext = ".json"
	if history_filename[-len(ext):] != ext:
		raise Exception("[ezRL error] history_filenameは拡張子が「%s」であることが想定されています"%ext)
	save_dir = "%s/"%(history_filename[:-len(ext)])
	if os.path.exists(save_dir) is True:
		raise Exception("[ezRL error] 保存先フォルダがすでに存在します")
	os.mkdir(save_dir)
	return save_dir

# 棋譜の画像化 (デフォルトは入力棋譜と同じディレクトリに置かれる)
def history2img(
	history_filename,	# 棋譜ファイル名
	game,	# ゲーム
	output_file_flag = True,	# ファイル出力フラグ
	file_out_ratio = 1.0
):
	# 棋譜ファイル読み込み
	with open(history_filename, "r", encoding = "utf-8") as f:
		history_ls = json.load(f)
	# 画像化
	img_ls = []
	for step in history_ls:
		state = step["state"]
		img = game.human_obs(state)
		img_ls.append(img)
	# ファイル出力フラグ判断
	if output_file_flag is True:
		# 保存先フォルダ作成
		save_dir = make_save_dir(history_filename)
		# 画像ファイル出力
		for idx, img in enumerate(img_ls):
			img_filename = "%s/%d.png"%(save_dir, idx)
			rout._raw_save_img(img, img_filename, ratio = file_out_ratio)	# pltを使ったcv2形式の画像保存
	return img_ls




# DQNの行動決定 [policy.py]

import sys
import random
import numpy as np

def arg_max(arg_ls):
	# インデックス付きリストの作成
	idx_e_ls = list(enumerate(arg_ls))
	# 最大値のインデックスを取得
	idx = max(idx_e_ls, key = lambda x: x[1])[0]
	return idx

# 1枚の画像に対する予測
def one_pred(obs_img, qnet):
	test_x = np.array([obs_img])
	raw_result = qnet.predict(test_x)
	q_ls = raw_result[0]
	return q_ls

# DQNの行動決定 (eps greedy戦略) [policy.py]
def eps_greedy_policy(obs_img, qnet, action_ls, eps):
	# Epsilon Greedyにより、一定確率でランダムな行動を選ぶ
	if random.random() < eps:
		# ランダムに行動を選ぶ
		selected_a = random.choice(action_ls)
	else:
		q_ls = one_pred(obs_img, qnet)	# 1枚の画像に対する予測
		selected_a = action_ls[arg_max(q_ls)]	# Q値が最も高いactionを選ぶ
	return selected_a




import sys
import random
from sout import sout
from relpath import add_import_path
add_import_path("../")
# Catcherゲーム [catcher_game]
from catcher_game import game_step, all_actions, human_obs, gen_init_state

# ランダム選択AI
class RandomPlayer:
	# 初期化処理
	def __init__(self, all_action_ls,
		game_func = None, ai_obs = None):
		self.all_action_ls = all_action_ls
	# 行動決定
	def think(self, state = None, reward = None):
		action = random.choice(self.all_action_ls)
		return action

# 動作確認・使用例
if __name__ == "__main__":
	# 全てのアクション
	ai = RandomPlayer(all_actions({}), game_func = None, ai_obs = None) # ランダム選択AI
	# state, actionの初期化
	action = "initial_action"
	state = gen_init_state(game_params = {})
	# ゲーム進行
	while state["finished"] is False:
		state, reward = game_step(state, action)
		action = ai.think(state, reward) # 行動決定
		# 人間用の観測関数 [catcher_game]
		human_obs(state, waitKey_msec = 200)



import sys
import ezpip
ezRL = ezpip.load_develop("ezRL", "../"*3)

history_filename = input("history_filename>")

# 棋譜の画像化 (デフォルトは入力棋譜と同じディレクトリに置かれる)
ezRL.history2img(
	history_filename,	# 棋譜ファイル名
	game = ezRL.catcher_game,	# ゲーム
	output_file_flag = True	# ファイル出力フラグ
)




# Triggerオブジェクト [trigger.py]

import sys

# triggerの指定方法が不正の例外を投げる
def raise_trigger_syntax_ex():
	raise Exception("triggerの指定方法が不正です")

# step, finishedから選ぶ
def switch_history(step_cnt_history, finished_cnt_history, config_type):
	if config_type == "step_n":
		return step_cnt_history
	elif config_type == "finished_n":
		return finished_cnt_history
	else:
		raise_trigger_syntax_ex()	# triggerの指定方法が不正の例外を投げる

# nの倍数を超えた瞬間を検知するトリガー
def sequence_trigger(sequence, n):
	if len(sequence) < 2: return False
	pre, now = sequence[-2:]
	flag = pre//n != now//n
	return flag

# triggerの指定に従ってjudge_funcを生成 (["finished_n", 3]など)
def gen_judge_func(trigger_config):
	# 関数指定されている場合
	if type(trigger_config) == type(lambda:None):
		judge_func = trigger_config
		return judge_func
	# 型の正当性の判定
	elif type(trigger_config) != type([]): raise_trigger_syntax_ex()	# triggerの指定方法が不正の例外を投げる
	if len(trigger_config) != 2: raise_trigger_syntax_ex()	# triggerの指定方法が不正の例外を投げる
	# リスト指定の場合
	config_type, n = trigger_config	# 設定値の解釈
	def judge_func(step_cnt_history, finished_cnt_history):
		# step, finishedから選ぶ
		history = switch_history(step_cnt_history, finished_cnt_history, config_type)
		# nの倍数を初めて超えた瞬間を検知するトリガー
		flag = sequence_trigger(sequence = history, n = n)
		return flag
	return judge_func

# Triggerオブジェクト [trigger.py]
class Trigger:
	# Triggerオブジェクトの初期化 (指定方法: ["finished", 3]または関数指定) [trigger.py]
	def __init__(self, trigger_config):
		# 履歴
		self.step_cnt_history = []
		self.finished_cnt_history = []
		# triggerの指定に従ってjudge_funcを生成 (["finished_n", 3]など)
		self.judge_func = gen_judge_func(trigger_config)
	# triggerタイミングの判断 [trigger.py]
	def judge(self, step_cnt, finished_cnt):
		# historyの更新
		self.step_cnt_history.append(step_cnt)
		self.finished_cnt_history.append(finished_cnt)
		# judge関数の呼び出し
		return self.judge_func(self.step_cnt_history, self.finished_cnt_history)




# qnetの双子 (qnetと重みを時々同期する) [target_twin.py]

import sys

# qnetの双子 (qnetと重みを時々同期する) [target_twin.py]
class TargetTwin:
	# 初期化処理
	def __init__(self, original_net):
		self.original_net = original_net
		# 未学習時の予測関数 (双子の姉のやつ)
		self.original_untrained_pred_func = self.original_net.untrained_pred_func
		self.model = None
	# 予測
	def predict(self, test_x):
		if self.model is None:
			return self.original_untrained_pred_func(test_x)
		else:
			return self.model.predict(test_x)
	# weightを更新 (双子の姉のweightに基づく) [target_twin.py]
	def update_by_twin_weight(self):
		if self.original_net.is_trained() is True:	# 未学習判定 (1度でも学習されたかどうか)
			if self.model is None:
				self.model = self.original_net.clone_model()	# 複製体を作る (返り値はNormal_CNN型) [normal_cnn]
			else:
				self.model.weight_update(self.original_net)	# weightを複写する (別のNormal_CNNオブジェクトから) [normal_cnn]
		else:
			# 未初期化時
			pass	# 何もしない




import random
import numpy as np

# schedule_configの指定方法が不正の例外を投げる
def raise_schedule_config_syntax_ex():
	raise Exception("schedule_configの指定方法が不正です")

# 端部耳あり線形スケジューラー (start, end, nの指定に基づく)
def linear_ear_scheduler(cnt, start, end, n):
	# 無次元化cntを求める
	raw_cnt_ratio = cnt / n
	# 無次元化cntをクリッピング
	cnt_ratio = min(max(raw_cnt_ratio, 0), 1)
	# 配分して返す
	return start + cnt_ratio * (end - start)

# eps値のスケジュール管理 [utils.py]
class EpsAdmin:
	# 初期化処理
	def __init__(self, eps_schedule):
		# 誤った型・形式で指定されている場合の対処
		if type(eps_schedule) == type([]):
			if len(eps_schedule) != 4: raise_schedule_config_syntax_ex()	# schedule_configの指定方法が不正の例外を投げる
			self.n_type = eps_schedule[2]
		elif type(eps_schedule) == type(lambda:None):
			self.n_type = "function"	# そのまま通す
		else:
			raise_schedule_config_syntax_ex()	# schedule_configの指定方法が不正の例外を投げる
		# クラスの持ち物
		self.eps_schedule = eps_schedule
		self.update(step_cnt = 0, finished_cnt = 0)	# 初期値の設定
	# eps値算出根拠の更新
	def update(self, step_cnt, finished_cnt):
		self.step_cnt, self.finished_cnt = step_cnt, finished_cnt
	# eps値の取得
	def get_eps(self):
		if self.n_type == "function":
			schedule_func = self.eps_schedule
			return schedule_func(self.step_cnt, self.finished_cnt)
		elif self.n_type == "step_n":
			start, end, _, n = self.eps_schedule	# 指定を解釈
			return linear_ear_scheduler(self.step_cnt, start, end, n)	# 端部耳あり線形スケジューラー (start, end, nの指定に基づく)
		elif self.n_type == "finished_n":
			start, end, _, n = self.eps_schedule	# 指定を解釈
			return linear_ear_scheduler(self.finished_cnt, start, end, n)	# 端部耳あり線形スケジューラー (start, end, nの指定に基づく)
		else:
			raise Exception("[EpsAdmin error] 未定義のn_typeが指定されています")

# 未学習時における予測関数の生成 (numpy配列の型で生成)
def gen_rand_pred_func(action_n):
	# 未学習時における予測関数
	def rand_pred_func(test_x):
		return np.array([
			[random.random() for _ in range(action_n)]
			for _ in test_x
		])
	return rand_pred_func




# プレイ記録 [playlog.py]

import sys
from sout import souts
from collections import deque

# プレイ記録 [playlog.py]
class DQNPlaylog:
	# 初期化処理
	def __init__(self, buffer_size):
		# プレイ記録の格納場所
		self.data = deque(
			[{"now_finished": True}],
			maxlen = buffer_size
		)
	# playlog追記 [playlog.py]
	def append(self, selected_a, state, obs_img, reward):
		self.data.append({
			"prev_obs": self.__get_prev_log("now_obs"), "prev_action": self.__get_prev_log("now_action"),	# 一つ前Stepのログの要素を参照する [playlog.py]
			"now_obs": obs_img,	"now_action": selected_a,
			"now_reward": reward, "now_finished": state["finished"]
		})
	# 全レコードを取得 [playlog.py]
	def get_all_data(self):
		return list(self.data)	# リスト型にして返す
	# 文字列化
	def __str__(self):
		ret_str = "last data of playlog:\n"
		ret_str += souts(self.data[-1], None) + "\n"
		ret_str += "(playlog entry n = %d)\n"%len(self.data)
		return ret_str
	# 文字列化その2
	def __repr__(self):
		return str(self)
	# 一つ前Stepのログの要素を参照する [playlog.py]
	def __get_prev_log(self, key):
		if self.data[-1]["now_finished"] is True:
			return None
		return self.data[-1][key]




# デモ [demo.py]

import sys
# 複数エピソード実行
from ..episode import do_episodes

# デモ (DQN-catcher) [demo.py]
def dqn_catcher_demo():
	import resout as rout
	from relpath import add_import_path
	add_import_path("../")
	import catcher_game as game	# Catcherゲーム [catcher_game]
	from DQN_Agent import DQN_Agent	# Deep Q Network AI [DQN_Agent]
	rout.set_save_dir("./ezRL_demo_output/")	# 保存パスの設定 [resout]
	train_ai = DQN_Agent(action_ls = game.all_actions({}), ai_obs = game.ai_obs) # Deep Q Network AI
	do_episodes(game, train_ai, game_params_ls = [{} for _ in range(700)], save_reward_ls = True)	# 複数エピソード実行
	test_ai = train_ai.gen_test()	# テスト用プレーヤーを生成 [DQN_Agent]
	do_episodes(game, test_ai, game_params_ls = [{} for _ in range(3)], save_history = True)	# 複数エピソード実行
	print("demo finished! (results are in \"./ezRL_demo_output/\".)")

# デモ (Random-catcher) [demo.py]
def random_catcher_demo():
	import resout as rout
	from relpath import add_import_path
	add_import_path("../")
	import catcher_game as game	# Catcherゲーム [catcher_game]
	from debug_players import RandomPlayer	# ランダムAI
	rout.set_save_dir("./ezRL_demo_output/")	# 保存パスの設定 [resout]
	train_ai = RandomPlayer(action_ls = game.all_actions({}), ai_obs = game.ai_obs)	# ランダムAI
	do_episodes(game, train_ai, game_params_ls = [{} for _ in range(30)], save_reward_ls = True)	# 複数エピソード実行
	test_ai = train_ai.gen_test()	# テスト用プレーヤーを生成 [DQN_Agent]
	do_episodes(game, test_ai, game_params_ls = [{} for _ in range(5)], save_history = True)	# 複数エピソード実行
	print("demo finished! (results are in \"./ezRL_demo_output/\".)")

# デモ (Random-catcher-PeriodicTest) [demo.py]
def random_catcher_periodicTest_demo():
	import resout as rout
	from relpath import add_import_path, rel2abs
	add_import_path("../")
	import catcher_game as game	# Catcherゲーム [catcher_game]
	from debug_players import RandomPlayer	# ランダムAI
	rout.set_save_dir("./ezRL_demo_output/")	# 保存パスの設定 [resout]
	# 定期テスト
	def rcpd_periodicTest(train_ai, episode_idx):
		if episode_idx%5 != (-1%5): return
		test_ai = train_ai.gen_test()	# テスト用プレーヤーを生成 [DQN_Agent]
		prob_name_ls = [str(i) for i in range(2)]
		do_episodes(game, test_ai, save_history = True,
			episode_label_ls = ["episode_idx=%d,pt_name=%s"%(episode_idx, name) for name in prob_name_ls],
			game_params_ls = [{"specified_stage": rel2abs("./handmade_stages/catcher_game/%s.json"%name)} for name in prob_name_ls]
		)
	train_ai = RandomPlayer(action_ls = game.all_actions({}), ai_obs = game.ai_obs)	# ランダムAI
	do_episodes(game, train_ai, game_params_ls = [{} for _ in range(30)], callback = rcpd_periodicTest, save_reward_ls = True)	# 複数エピソード実行
	test_ai = train_ai.gen_test()	# テスト用プレーヤーを生成 [DQN_Agent]
	do_episodes(game, test_ai, game_params_ls = [{} for _ in range(5)], save_history = True)	# 複数エピソード実行
	print("demo finished! (results are in \"./ezRL_demo_output/\".)")

# デモ [demo.py]
def demo(demo_name = "DQN-catcher"):
	if demo_name == "DQN-catcher":
		dqn_catcher_demo()	# デモ (DQN-catcher) [demo.py]
	elif demo_name == "Random-catcher":
		random_catcher_demo()	# デモ (Random-catcher) [demo.py]
	elif demo_name == "Random-catcher-PeriodicTest":
		random_catcher_periodicTest_demo()	# デモ (Random-catcher-PeriodicTest) [demo.py]
	else:
		raise Exception("[ezRL error] invalid demo name.")




# シンプルなCNN [normal_cnn]

import sys
import numpy as np
from relpath import add_import_path
from sklearn import datasets, model_selection, metrics
add_import_path("../")
# シンプルなCNN [normal_cnn]
from normal_cnn import Normal_CNN

# 【動作確認・使用例】
if __name__ == '__main__':
	# CNNの構成の設定
	nn_param = {
		"conv_layer_info": [
			{"n_filter": 8, "kernel_size": (5, 5)},
			{"n_filter": 8, "kernel_size": (5, 5)},
			{"n_filter": 8, "kernel_size": (5, 5)},
		],
		"dense_layer_info": [
			{"unit_n": 8, "activation": "relu"},
		],
		"output_activation": "softmax",
	}
	# 出力層のユニット数
	n_class = 10
	# モデルの定義
	cnn = Normal_CNN(n_class, nn_param, loss = "categorical_crossentropy")	# シンプルなCNN [normal_cnn]

	# 軽量数字データセット
	digits = datasets.load_digits()
	img_shape = (8, 8, 1)
	y = digits["target"]
	X = np.array([flat_img.reshape(img_shape) for flat_img in digits["data"]])
	# 学習データ、テストデータの分離
	splited = model_selection.train_test_split(
											X, 
											y, 
											test_size = 0.33)
	X_train, X_test, y_train, y_test = splited
	# モデルの学習・予測
	from keras.utils.np_utils import to_categorical
	categ_train = to_categorical(y_train, num_classes = n_class)	# one-hot encoding
	cnn.fit(X_train, categ_train)	# 学習
	categorical_pred = cnn.predict(X_test)	# 予測
	y_pred = [np.argmax(one_pred) for one_pred in categorical_pred]
	# 予測結果の表示
	accuracy = metrics.accuracy_score(y_test, y_pred)
	cof_mat = metrics.confusion_matrix(y_test, y_pred)
	print("Accuracy: %.2f"%accuracy)
	print("Confusion matrix:")
	print(cof_mat)




# debug用プレーヤー [debug_players]
# 【動作確認 / 使用例】

import sys
import resout as rout
import ezpip
ezRL = ezpip.load_develop("ezRL", "../../../")
game = ezRL.catcher_game

def human_obs_func(state):
	rout.save_img(game.human_obs(state), ratio = 2)	# 画像の保存 [resout]

rout.set_save_dir("./results/")	# 保存パスの設定 [resout]
h_player = ezRL.HumanPlayer(action_ls = game.all_actions({}), human_obs_func = human_obs_func)	# 人間プレーヤー [ezRL]
raise Exception("do_episodes()関数の新仕様に未対応")
ezRL.do_episodes(game, h_player, game_params = {}, episode_n = 1)	# 複数エピソード実行
